#include <string>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#ifndef _MSC_VER
#include <sys/wait.h>
#include <pty.h>
#include <utmp.h>
#include <SDL2/SDL.h>
#else
#include <SDL.h>
#endif
#include "Terminal.h"
#include "fonts/cp437.h"
#include "fonts/potnoodle.h"
#include "fonts/topazplus.h"
#include "fonts/mkplus.h"
#include "fonts/microknight.h"
#include "fonts/topaz.h"
#include "fonts/mosoul.h"

extern ssh_channel chan;
extern SDL_Thread *dlthread;

struct zmodem_info {
    int *ptr;
    char *data;
    int len;
    char *uploadfile;
};

struct zmodem_info zm_inf;

extern SDL_Surface *loadImage(unsigned char *iaddr, int ilen);

int do_zmodem_download(void *ptr);
int do_zmodem_upload(void *ptr);

void zmodem_upload(char *filename) {
    if (zm_inf.uploadfile != NULL) {
        free(zm_inf.uploadfile);
    }

    zm_inf.uploadfile = strdup(filename);

    dlthread = SDL_CreateThread(do_zmodem_upload, "ZMODEM", &zm_inf);
}


#ifndef _MSC_VER
int ttySetRaw(int fd, struct termios *prevTermios) {
    struct termios t;

    if (tcgetattr(fd, &t) == -1)
        return -1;

    if (prevTermios != NULL)
        *prevTermios = t;

    t.c_lflag &= ~(ICANON | ISIG | IEXTEN | ECHO);
    t.c_iflag &= ~(BRKINT | ICRNL | IGNBRK | IGNCR | INLCR | INPCK | ISTRIP | IXON | PARMRK);
    t.c_oflag &= ~OPOST;
    t.c_cc[VMIN] = 1;
    t.c_cc[VTIME] = 0;

    if (tcsetattr(fd, TCSAFLUSH, &t) == -1)
        return -1;

    return 0;
}
#endif
Terminal::Terminal()
{
    //ctor
    Uint32 rmask, gmask, bmask, amask;

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000;
    gmask = 0x00ff0000;
    bmask = 0x0000ff00;
    amask = 0x000000ff;
#else
    rmask = 0x000000ff;
    gmask = 0x0000ff00;
    bmask = 0x00ff0000;
    amask = 0xff000000;
#endif

    this->screen = SDL_CreateRGBSurface(SDL_SWSURFACE, 640, 480, 32, rmask, gmask, bmask, amask);
    this->screenbacker = SDL_CreateRGBSurface(SDL_SWSURFACE, 640, 400, 32, rmask, gmask, bmask, amask);
    this->screencopy = SDL_CreateRGBSurface(SDL_SWSURFACE, 640, 400, 32, rmask, gmask, bmask, amask);
    this->font_bmp = loadImage((unsigned char *)CP437_IMAGE_DATA, sizeof(CP437_IMAGE_DATA));
    this->pot_noodle_bmp = loadImage((unsigned char *)POTNOODLE_IMAGE_DATA, sizeof(POTNOODLE_IMAGE_DATA));
    this->topaz_plus_bmp = loadImage((unsigned char *)TOPAZPLUS_IMAGE_DATA, sizeof(TOPAZPLUS_IMAGE_DATA));
    this->mk_plus_bmp = loadImage((unsigned char *)MKPLUS_IMAGE_DATA, sizeof(MKPLUS_IMAGE_DATA));
    this->mk_bmp = loadImage((unsigned char *)MK_IMAGE_DATA, sizeof(MK_IMAGE_DATA));
    this->topaz_bmp = loadImage((unsigned char *)TOPAZ_IMAGE_DATA, sizeof(TOPAZ_IMAGE_DATA));    
    this->mosoul_bmp = loadImage((unsigned char *)MOSOUL_IMAGE_DATA, sizeof(MOSOUL_IMAGE_DATA));   
    this->color = 7;
    this->bgcolor = 0;
    this->cur_col = 0;
    this->cur_row = 0;
    this->state = 0;
    this->bold = 0;
    this->save_col = 0;
    this->save_row = 0;
    this->vt320 = 0;
    this->cursor = 1;
    this->zmodem_detect = 0;
    this->zmodem = NULL;
    this->curfont = 0;

    zm_inf.data = NULL;
    zm_inf.uploadfile = NULL;
}

void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel) {
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        *p = pixel;
        break;

    case 2:
        *(Uint16 *)p = pixel;
        break;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (pixel >> 16) & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = pixel & 0xff;
        } else {
            p[0] = pixel & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = (pixel >> 16) & 0xff;
        }
        break;

    case 4:
        *(Uint32 *)p = pixel;
        break;
    }
}

Uint32 getpixel(SDL_Surface *surface, int x, int y) {
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        return *p;
        break;

    case 2:
        return *(Uint16 *)p;
        break;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;
        break;

    case 4:
        return *(Uint32 *)p;
        break;

    default:
        return 0;       /* shouldn't happen, but avoids warnings */
    }
}

void Terminal::scrollUp() {
    SDL_Rect src, dest;
    src.x = 0;
    src.y = 16;
    src.w = 640;
    src.h = 384;
    dest.x = 0; 
    dest.y = 0;
    dest.w = 640;
    dest.h = 384;
    SDL_FillRect(screencopy, 0, 0xff000000);
    SDL_BlitSurface(screenbacker, &src, screencopy, &dest);
    src.x = 0;
    src.y = 0;
    src.w = 640;
    src.h = 400;
    dest.x = 0; 
    dest.y = 0;
    dest.w = 640;
    dest.h = 400;   
    SDL_FillRect(screenbacker, 0, 0xff000000);     
    SDL_BlitSurface(screencopy, &src, screenbacker, &dest);
}

void Terminal::drawChar(unsigned char c) {
    int x;
    int y;
    SDL_Rect src, dest;

    if (cur_col == 80) {
        cur_col = 0;
        cur_row++;
    }
    if (cur_row >= 25) {
        // scroll_up();
        scrollUp();
        cur_row = 24;
    }  

    src.x = (c % 32) * 8;
    src.y = (c / 32) * 16;
    src.w = 8;
    src.h = 16;

    dest.x = cur_col * 8;
    dest.y = cur_row * 16;
    dest.w = 8;
    dest.h = 16;

    switch(curfont) {
        case 0:
            SDL_BlitSurface(font_bmp, &src, screenbacker, &dest);
            break;
        case 1:
            SDL_BlitSurface(pot_noodle_bmp, &src, screenbacker, &dest);
            break;
        case 2:
            SDL_BlitSurface(topaz_plus_bmp, &src, screenbacker, &dest);
            break;
        case 3:
            SDL_BlitSurface(mk_plus_bmp, &src, screenbacker, &dest);
            break;
        case 4:
            SDL_BlitSurface(mk_bmp, &src, screenbacker, &dest);
            break;
        case 5:
            SDL_BlitSurface(topaz_bmp, &src, screenbacker, &dest);            
            break;
        case 6:
            SDL_BlitSurface(mosoul_bmp, &src, screenbacker, &dest);            
            break;            
    }
    

    SDL_LockSurface(screenbacker);

    for (x = dest.x; x< dest.x + dest.w; x++) {
        for (y = dest.y; y < dest.y + dest.h; y++) {
            if (getpixel(screenbacker, x, y) == 0xff000000) {
                putpixel(screenbacker, x, y, colours[color]);
            } else {
                putpixel(screenbacker, x, y, colours[bgcolor]);
            }
        }
    }
    SDL_UnlockSurface(screenbacker);
    cur_col++;
  
}

void Terminal::twrite(unsigned char *c, int len)
{
    int i, j, k;
    SDL_Rect dest;
    SDL_Rect src;
    char buf[10];


    for (i=0;i<len;i++) {
        if (state == 0) {
            if (c[i] == 0x1b) {
                state = 1;
            } else {
                if (c[i] == '\r') {
                    cur_col=0;
                } else if (c[i] == '\n') {
                    cur_row++;
                    if (cur_row == 25) {
                        scrollUp();
                        cur_row--;
                    }
                } else if (c[i] == 8) {
                    cur_col--;
                    if (cur_col < 0) {
                        cur_col = 0;
                    }
                } else {
                    if (c[i] == 24 && zmodem_detect == 0) {
                        zmodem_detect = 1;
                    } else if (c[i] == 'B' && zmodem_detect == 1) {
                        zmodem_detect = 2;
                    } else if (c[i] == '0' && zmodem_detect == 2) {
                        zmodem_detect = 3;
                    } else if (c[i] == '0' && zmodem_detect == 3) {
                        if (zmodem != NULL) {
                            *zmodem = 1;
                            if (zm_inf.data != NULL) {
                                free(zm_inf.data);
                            }
                            zm_inf.data = (char *)malloc((len - i) * (sizeof(char)));
                            memcpy(zm_inf.data, &c[i], len - i);
                            zm_inf.len = len -i;
                            zm_inf.ptr = zmodem;
                            dlthread = SDL_CreateThread(do_zmodem_download, "ZMODEM", &zm_inf);
                        }
                        zmodem_detect = 0;
                        return;
                    } else if (c[i] == '1' && zmodem_detect == 3) {
                        *zmodem = 2;
                        if (zm_inf.data != NULL) {
                            free(zm_inf.data);
                        }
                        zm_inf.data = (char *)malloc((len - i) * (sizeof(char)));
                        memcpy(zm_inf.data, &c[i], len - i);
                        zm_inf.len = len -i;
                        zm_inf.ptr = zmodem;
                        zmodem_detect = 0;
                        return;                       
                    } else {
                        zmodem_detect = 0;
                    }
                    drawChar(c[i]);
                }
            }
        } else if (state == 1) {
            if (c[i] == '[') {
                state = 2;
                continue;
            } else {
                state = 0;
                printf("Not CSI %d\n", c[i]);
                continue;
            }
        } else if (state == 2) {
            param_count = 0;
            for (j=0;j<16;j++) {
                params[j] = 0;
            }
            state = 3;
        }

        if (state == 3) {
            if (c[i] == ';') {
				if (param_count < 15) {
					param_count++;
				}
				continue;
			} else if (c[i] >= '0' && c[i] <= '9') {
				if (!param_count) param_count = 1;
				params[param_count-1] = params[param_count-1] * 10 + (c[i] - '0');
				continue;
			} else {
				state = 4;
			}
        }

        if (state == 4) {
            switch (c[i]) {
                case 'H':
                case 'f':
                    if (params[0]) params[0]--;
                    if (params[1]) params[1]--;

                    cur_col = params[1];
                    cur_row = params[0];
                    if (cur_col > 79) {
                        cur_col = 79;
                    }

                    if (cur_row > 24) {
                        cur_row = 24;
                    }
                    state = 0;
                    break;
                case 'A':
                    if (param_count > 0) {
                        cur_row -= params[0];
                    } else {
                        cur_row--;
                    }

                    if (cur_row < 0) {
                        cur_row = 0;
                    }
                    state = 0;
                    break;
                case 'B':
                    if (param_count > 0) {
                        cur_row += params[0];
                    } else {
                        cur_row++;
                    }

                    if (cur_row > 24) {
                        cur_row = 24;
                    }
                    state = 0;
                    break;
				case 'C':
					if (param_count > 0) {
                        cur_col += params[0];
					} else {
                        cur_col++;
					}

                    if (cur_col > 79) {
                        cur_col = 79;
                    }
					state = 0;
					break;
            	case 'D':
					if (param_count > 0) {
						cur_col -= params[0];
					} else {
						cur_col--;
					}

                    if (cur_col < 0) {
                        cur_col = 0;
                    }

					state = 0;
					break;
                case 's':
                    save_col = cur_col;
                    save_row = cur_row;

                    state = 0;
                    break;
                case 'u':
                    cur_col = save_col;
                    cur_row = save_row;

                    state = 0;
                    break;
                case 'm':
                    for (j=0;j<param_count;j++) {
                        switch (params[j]) {
                            case 0:
                                color = 7;
                                bgcolor = 0;
                                bold = 0;
                                break;
                            case 1:
                                bold = 1;
                                {
                                    switch(color) {
                                        case 0:
                                            color = 8;
                                            break;
                                        case 4:
                                            color = 12;
                                            break;
                                        case 2:
                                            color = 10;
                                            break;
                                        case 6:
                                            color = 14;
                                            break;
                                        case 1:
                                            color = 9;
                                            break;
                                        case 5:
                                            color = 13;
                                            break;
                                        case 3:
                                            color = 11;
                                            break;
                                        case 7:
                                            color = 15;
                                            break;
                                        default:
                                            break;
                                    }
                                    
                                }
                                break;
                            case 2:
                                bold = 0;
                                {
                                    switch(color) {
                                        case 8:
                                            color = 0;
                                            break;
                                        case 12:
                                            color = 4;
                                            break;
                                        case 10:
                                            color = 2;
                                            break;
                                        case 14:
                                            color = 6;
                                            break;
                                        case 9:
                                            color = 1;
                                            break;
                                        case 13:
                                            color = 5;
                                            break;
                                        case 11:
                                            color = 3;
                                            break;
                                        case 15:
                                            color = 7;
                                            break;
                                        default:
                                            break;
                                    }
                                    
                                }
                                break;
                            case 30:
                                if (bold) {
                                    color = 8;
                                } else {
                                    color = 0;
                                }
                                break;
                            case 31:
                                if (bold) {
                                    color = 12;
                                } else {
                                    color = 4;
                                }
                                break;
                            case 32:
                                if (bold) {
                                    color = 10;
                                } else {
                                    color = 2;
                                }
                                break;
                            case 33:
                                if (bold) {
                                    color = 14;
                                } else {
                                    color = 6;
                                }
                                break;
                            case 34:
                                if (bold) {
                                    color = 9;
                                } else {
                                    color = 1;
                                }
                                break;
                            case 35:
                                if (bold) {
                                    color = 13;
                                } else {
                                    color = 5;
                                }
                                break;
                            case 36:
                                if (bold) {
                                    color = 11;
                                } else {
                                    color = 3;
                                }
                                break;
                            case 37:
                                if (bold) {
                                    color = 15;
                                } else {
                                    color = 7;
                                }
                                break;
                            case 40:
                                bgcolor = 0;
                                break;
                            case 41:
                                bgcolor = 4;
                                break;
                            case 42:
                                bgcolor = 2;
                                break;
                            case 43:
                                bgcolor = 6;
                                break;
                            case 44:
                                bgcolor = 1;
                                break;
                            case 45:
                                bgcolor = 5;
                                break;
                            case 46:
                                bgcolor = 3;
                                break;
                            case 47:
                                bgcolor = 7;
                                break;                                
                        }
                    }
                    state = 0;
                    break;
                case 'J':
					if (params[0] == 0) {
                        dest.x = 0;
                        dest.y = 16 * cur_row;
                        dest.w = 80 * 16;
                        dest.h = (25 - cur_row) * 16;
                        SDL_FillRect(screenbacker, &dest, bgcolor);
					} else if (params[0] == 1) {
                        dest.x = 0;
                        dest.y = 0;
                        dest.w = 80 * 16;
                        dest.h = cur_row * 16;
                        SDL_FillRect(screenbacker, &dest, bgcolor);
					} else if (params[0] == 2) {
						SDL_FillRect(screenbacker, 0, bgcolor);
						cur_col = 0;
						cur_row = 0;
					}
					state = 0;                
                    break;
                case 'K':
					if (params[0] == 0) {
						j = cur_col;
						for (k=j;k<80;k++) {
							drawChar(' ');
						}
						cur_col = j;
					} else if (params[0] == 1) {
						j = cur_col;
						cur_col = 0;
						for (k=0;k<j;k++) {
							drawChar(' ');
						}
						cur_col = j;
					} else if (params[0] == 2) {
						j = cur_col;
						for (k=0;k<80;k++) {
							drawChar(' ');
						}
						cur_col = j;
					}
					state = 0;
                    break;
                case 'n':
                    // report cursor position
                    if (chan != NULL) {
                        snprintf(buf, 10, "\x1b[%d;%dR", cur_row, cur_col);
                        ssh_channel_write(chan, buf, strlen(buf));
                    }
                    break;
                case '?':
                    vt320 = 0;
                    state = 5;
                    break;
                case 27:
                    state = 1;
                    break;
                case ' ':
                    state = 7;
                    break;
                case '*':
                    state = 8;
                    break;
                default:
                    printf ("Unknown ansi code %d\n", c[i]);
                    state = 0;
                    break;
            }
        } else if (state == 5) {
            if (c[i] >= '0' && c[i] <= '9') {
                vt320 = vt320 * 10 + (c[i] - '0');
                continue;
            } else {
                state = 6;
            }
        } else if (state == 7) {
            if (c[i] == 'D') {
                switch(params[1]) {
                    case 37:
                        curfont = 1;
                        break;
                    case 40:
                        curfont = 2;
                        break;
                    case 39:
                        curfont = 3;
                        break;
                    case 41:
                        curfont = 4;
                        break;
                    case 42:
                        curfont = 5;
                        break;
                    case 38:
                        curfont = 6;
                        break;
                    default:
                        curfont = 0;
                        break;
                }
                printf("Got font change request %d\n", params[1]);
            }
            state = 0;
        } else if (state == 8) {
            if (c[i] == 'r') {
                printf("Got Speed Emulation Change Request\n");
            }
            state = 0;
        }

        if (state == 6) {
            if (vt320 == 25) {
                if (c[i] == 'h') {
                    cursor = 1;
                }
                if (c[i] == 'l') {
                    cursor = 0;
                }
            }
            state = 0;
        }

    
    }
    return;
}
#ifndef _MSC_VER
int do_zmodem_upload(void *ptr) {

    struct stat s;
    const char *path = getenv("PATH");
    char *path_copy = strdup(path);
    char *ptr2;
    char filename[PATH_MAX];
    char buffer[256];
    int found = 0;
    int len = 0;
    int d;
    struct zmodem_info *zi = (struct zmodem_info *)ptr;
    int *zmodem = zi->ptr;
    int master, slave;
    int status;
    pid_t pid;

    ptr2 = strtok(path_copy, ":");
    while (ptr2 != NULL) {
        snprintf(filename, PATH_MAX, "%s/sz", ptr2);
        if (stat(filename, &s) == 0) {
            found = 1;
            break;
        }
        snprintf(filename, PATH_MAX, "%s/lsz", ptr2);
        if (stat(filename, &s) == 0) {
            found = 1;
            break;
        }
        ptr2 = strtok(NULL, ":");
    }

    if (chan != NULL && found) {
            openpty(&master, &slave, NULL, NULL, NULL);

            ttySetRaw(master, NULL);
            ttySetRaw(slave, NULL);

            pid = fork();

            switch (pid) {
                case -1:
                    perror("fork");
                    //twrite((unsigned char *)"\r\nFORK Failed\r\n", 15);
                    break;
                case 0:
                    
                    close(master);

                    dup2(slave, 0);
                    dup2(slave, 1);		
                        
                    close(slave);

                    setsid();

                    ioctl(0, TIOCSCTTY, 1);

                    close(master);
                    execlp(filename, filename, "-v", zi->uploadfile, NULL);
                    exit(-1);
                    break;
                default:
                    // parent
                    close(slave);
                    
                    fcntl(master, F_SETFL, FNDELAY);

                    write(master, "**\030B01", 6);
                    write(master, zi->data, zi->len);

                    while(waitpid(pid, &status, WNOHANG) == 0) {
                        len = read(master, buffer, 256);
                        if (len > 0) {
                            ssh_channel_write(chan, buffer, len);
                        } 

                        len = ssh_channel_read_nonblocking(chan, buffer, 256, 0);
                        if (len > 0) {
                            write(master, buffer, len);
                        } else if (len < 0) {
                            kill(pid, SIGHUP);
                        }
                    }
                    close(master);
                    break;
            }
    }
    *zmodem = 0;
	return 0;
}

int do_zmodem_download(void *ptr) {

    struct stat s;
    const char *path = getenv("PATH");
    char *path_copy = strdup(path);
    char *ptr2;
    char filename[PATH_MAX];
    char buffer[256];
    int found = 0;
    int len = 0;
    int d;
    struct zmodem_info *zi = (struct zmodem_info *)ptr;
    int *zmodem = zi->ptr;
    int master, slave;
    int status;
    pid_t pid;

    ptr2 = strtok(path_copy, ":");
    while (ptr2 != NULL) {
        snprintf(filename, PATH_MAX, "%s/rz", ptr2);
        if (stat(filename, &s) == 0) {
            found = 1;
            break;
        }
        snprintf(filename, PATH_MAX, "%s/lrz", ptr2);
        if (stat(filename, &s) == 0) {
            found = 1;
            break;
        }
        ptr2 = strtok(NULL, ":");
    }

    if (chan != NULL && found) {
            openpty(&master, &slave, NULL, NULL, NULL);

            ttySetRaw(master, NULL);
            ttySetRaw(slave, NULL);

            pid = fork();

            switch (pid) {
                case -1:
                    perror("fork");
                    //twrite((unsigned char *)"\r\nFORK Failed\r\n", 15);
                    break;
                case 0:
                    
                    close(master);

                    dup2(slave, 0);
                    dup2(slave, 1);		
                        
                    close(slave);

                    setsid();

                    ioctl(0, TIOCSCTTY, 1);

                    close(master);
                    execlp(filename, filename, "-v", NULL);
                    exit(-1);
                    break;
                default:
                    // parent
                    close(slave);
                    
                    fcntl(master, F_SETFL, FNDELAY);

                    write(master, "**\030B00", 6);
                    write(master, zi->data, zi->len);

                    while(waitpid(pid, &status, WNOHANG) == 0) {
                        len = read(master, buffer, 256);
                        if (len > 0) {
                            ssh_channel_write(chan, buffer, len);
                        } 

                        len = ssh_channel_read_nonblocking(chan, buffer, 256, 0);
                        if (len > 0) {
                            write(master, buffer, len);
                        } else if (len < 0) {
                            kill(pid, SIGHUP);
                        }
                    }
                    close(master);
                    break;
            }
    }
    *zmodem = 0;
	return 0;
}
#else 

int do_zmodem_upload(void *ptr) {
	struct zmodem_info *zi = (struct zmodem_info *)ptr;
	int *zmodem = zi->ptr;
    HANDLE g_hChildStd_IN_Rd = NULL;
    HANDLE g_hChildStd_IN_Wr = NULL;
    HANDLE g_hChildStd_OUT_Rd = NULL;
    HANDLE g_hChildStd_OUT_Wr = NULL;

    BOOL bSuccess = FALSE;
    DWORD code;
    DWORD dwRead, dwWritten; 
    CHAR chBuf[4096];
	DWORD available;

    SECURITY_ATTRIBUTES saAttr; 
    PROCESS_INFORMATION piProcInfo; 
    STARTUPINFO siStartInfo;
    
    if (chan != NULL) {
        saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
        saAttr.bInheritHandle = TRUE; 
        saAttr.lpSecurityDescriptor = NULL;
        
        CreatePipe(&g_hChildStd_OUT_Rd, &g_hChildStd_OUT_Wr, &saAttr, 0);
        SetHandleInformation(g_hChildStd_OUT_Rd, HANDLE_FLAG_INHERIT, 0);
        CreatePipe(&g_hChildStd_IN_Rd, &g_hChildStd_IN_Wr, &saAttr, 0);
        SetHandleInformation(g_hChildStd_IN_Wr, HANDLE_FLAG_INHERIT, 0);

		ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
		siStartInfo.cb = sizeof(STARTUPINFO);
		siStartInfo.hStdOutput = g_hChildStd_OUT_Wr;
		siStartInfo.hStdInput = g_hChildStd_IN_Rd;
		siStartInfo.dwFlags |= STARTF_USESTDHANDLES;

        sprintf(chBuf, "sz -v -e %s", zi->uploadfile);

        CreateProcess(NULL, 
            chBuf,     // command line 
            NULL,          // process security attributes 
            NULL,          // primary thread security attributes 
            TRUE,          // handles are inherited 
            0,             // creation flags 
            NULL,          // use parent's environment 
            NULL,          // use parent's current directory 
            &siStartInfo,  // STARTUPINFO pointer 
            &piProcInfo);  // receives PROCESS_INFORMATION 
        
		WriteFile(g_hChildStd_IN_Wr, "**\030B01", 6, &dwWritten, NULL);
		WriteFile(g_hChildStd_IN_Wr, zi->data, zi->len, &dwWritten, NULL);

        do {
			if (PeekNamedPipe(g_hChildStd_OUT_Rd, NULL, 0, NULL, &available, NULL)) {
				if (available > 0) {
					bSuccess = ReadFile(g_hChildStd_OUT_Rd, chBuf, 4096, &dwRead, NULL);
					if (dwRead > 0) {
						ssh_channel_write(chan, chBuf, dwRead);
					}
				}
			} 
            dwRead = ssh_channel_read_nonblocking(chan, chBuf, 4096, 0);
            if (dwRead > 0) {
                bSuccess = WriteFile(g_hChildStd_IN_Wr, chBuf, dwRead, &dwWritten, NULL);
            } else if (dwRead < 0) {
                TerminateProcess(piProcInfo.hProcess, 1);
                WaitForSingleObject(piProcInfo.hProcess, INFINITE);
            }

            GetExitCodeProcess(piProcInfo.hProcess, &code);
        } while (code == STILL_ACTIVE);

        CloseHandle(piProcInfo.hProcess);
        CloseHandle(piProcInfo.hThread);
    }
    *zmodem = 0;

	return 0;
}

int do_zmodem_download(void *ptr) {
	struct zmodem_info *zi = (struct zmodem_info *)ptr;
	int *zmodem = zi->ptr;
    HANDLE g_hChildStd_IN_Rd = NULL;
    HANDLE g_hChildStd_IN_Wr = NULL;
    HANDLE g_hChildStd_OUT_Rd = NULL;
    HANDLE g_hChildStd_OUT_Wr = NULL;

    BOOL bSuccess = FALSE;
    DWORD code;
    DWORD dwRead, dwWritten; 
    CHAR chBuf[4096];
	DWORD available;

    SECURITY_ATTRIBUTES saAttr; 
    PROCESS_INFORMATION piProcInfo; 
    STARTUPINFO siStartInfo;
    
    if (chan != NULL) {
        saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
        saAttr.bInheritHandle = TRUE; 
        saAttr.lpSecurityDescriptor = NULL;
        
        CreatePipe(&g_hChildStd_OUT_Rd, &g_hChildStd_OUT_Wr, &saAttr, 0);
        SetHandleInformation(g_hChildStd_OUT_Rd, HANDLE_FLAG_INHERIT, 0);
        CreatePipe(&g_hChildStd_IN_Rd, &g_hChildStd_IN_Wr, &saAttr, 0);
        SetHandleInformation(g_hChildStd_IN_Wr, HANDLE_FLAG_INHERIT, 0);

		ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
		siStartInfo.cb = sizeof(STARTUPINFO);
		siStartInfo.hStdOutput = g_hChildStd_OUT_Wr;
		siStartInfo.hStdInput = g_hChildStd_IN_Rd;
		siStartInfo.dwFlags |= STARTF_USESTDHANDLES;

        CreateProcess(NULL, 
            "rz -v -e",     // command line 
            NULL,          // process security attributes 
            NULL,          // primary thread security attributes 
            TRUE,          // handles are inherited 
            0,             // creation flags 
            NULL,          // use parent's environment 
            NULL,          // use parent's current directory 
            &siStartInfo,  // STARTUPINFO pointer 
            &piProcInfo);  // receives PROCESS_INFORMATION 
        
		WriteFile(g_hChildStd_IN_Wr, "**\030B00", 6, &dwWritten, NULL);
		WriteFile(g_hChildStd_IN_Wr, zi->data, zi->len, &dwWritten, NULL);

        do {
			if (PeekNamedPipe(g_hChildStd_OUT_Rd, NULL, 0, NULL, &available, NULL)) {
				if (available > 0) {
					bSuccess = ReadFile(g_hChildStd_OUT_Rd, chBuf, 4096, &dwRead, NULL);
					if (dwRead > 0) {
						ssh_channel_write(chan, chBuf, dwRead);
					}
				}
			} 
            dwRead = ssh_channel_read_nonblocking(chan, chBuf, 4096, 0);
            if (dwRead > 0) {
                bSuccess = WriteFile(g_hChildStd_IN_Wr, chBuf, dwRead, &dwWritten, NULL);
            } else if (dwRead < 0) {
                TerminateProcess(piProcInfo.hProcess, 1);
                WaitForSingleObject(piProcInfo.hProcess, INFINITE);
            }

            GetExitCodeProcess(piProcInfo.hProcess, &code);
        } while (code == STILL_ACTIVE);

        CloseHandle(piProcInfo.hProcess);
        CloseHandle(piProcInfo.hThread);
    }
    *zmodem = 0;

	return 0;
}
#endif

void Terminal::setZmodem(int *z) {
    zmodem = z;
}


SDL_Surface *Terminal::getSurface() {
    SDL_Rect src, dest;
    src.x = 0;
    src.y = 0;
    src.w = 640;
    src.h = 400;

    dest.x = 0;
    dest.y = 0;
    dest.w = 640;
    dest.h = 400;
    SDL_FillRect(screen, &src, 0xff000000);
    SDL_BlitSurface(screenbacker, &src, screen, &dest);

    if (cursor) {
        src.x = cur_col * 8;
        src.y = cur_row * 16 + 14;
        src.w = 8;
        src.h = 2;
        
        SDL_FillRect(screen, &src, 0xffffffff);
    }
    return screen;
}

Terminal::~Terminal()
{
    //dtor
    SDL_FreeSurface(screen);
    SDL_FreeSurface(screenbacker);
    SDL_FreeSurface(screencopy);
    
    SDL_FreeSurface(font_bmp);
    SDL_FreeSurface(pot_noodle_bmp);
    SDL_FreeSurface(topaz_plus_bmp);
    SDL_FreeSurface(topaz_bmp);
    SDL_FreeSurface(mk_bmp);
    SDL_FreeSurface(mk_plus_bmp);
    SDL_FreeSurface(mosoul_bmp);
}
